/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_ROW_PROXY_HPP
#define DATAFRAME_ROW_PROXY_HPP 1

#include <boost/operators.hpp>
#include <observer_ptr.hpp>

namespace dataframe
{

  template < typename Dataframe >
  class row_proxy
    : boost::totally_ordered<row_proxy<Dataframe>>
  {
    using self = row_proxy;
    using pointer = observer_ptr<Dataframe const>;
    using size_type = typename Dataframe::size_type;
    using header_type = typename Dataframe::header_type;

  public:
    row_proxy () = default;
    
    explicit row_proxy (Dataframe const& dataframe)
      : p_(make_observer(dataframe))
    {}

    row_proxy (Dataframe const& dataframe, size_type row)
      : p_(make_observer(dataframe))
      , row_(row)
    {}

    auto at (size_type j) const { return dataframe().element(row_,j); }
    auto at (header_type header) const { return dataframe().element(row_,header); }

    template < typename T >
    auto at (size_type j) const { return dataframe().template element<T>(row_,j); }

    template < typename T >
    auto at (header_type header) const { return dataframe().template element<T>(row_,header); }

    auto operator () (size_type j) const { return at(j); }
    auto operator () (header_type header) const { return at(header); }

    auto row () const { assert(p_); return row_; }

    operator size_type () const { return row(); }

    friend bool operator == (self left, self right)
    {
      return
	(left.is_end() && right.is_end())
	|| (!left.is_end() && !right.is_end()
	    && left.p_ == right.p_
	    && left.row_ == right.row_);
    }

    void increment () { ++row_; }
    void decrement () { --row_; }
    template < typename N >
    auto advance (N n) { row_ += n; return *this; }
    auto distance_to (self other) { assert(other.p_ == p_); return other.row_ - row_; }

  private:
    bool is_end () const { return !p_ || row_ == dataframe().size(); }
    auto& dataframe () { return *p_; }
    auto& dataframe () const { return *p_; }

  private:
    pointer p_;
    size_type row_ = 0;
  };

  template < typename Dataframe >
  auto make_row_proxy (Dataframe const& dataframe)
  {
    using row_proxy_type = row_proxy<Dataframe>;
    return row_proxy_type(dataframe);
  }

  template < typename Dataframe, typename Row >
  auto make_row_proxy (Dataframe const& dataframe, Row row)
  {
    using row_proxy_type = row_proxy<Dataframe>;
    return row_proxy_type(dataframe,row);
  }

} // namespace dataframe

#endif
