/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_LATTICE_HPP
#define GRAPH_MODEL_GRAPH_LATTICE_HPP 1

#include <algorithm>
#include <type_traits>
#include <boost/array.hpp>
#include <boost/graph/grid_graph.hpp>
#include <boost/property_map/shared_array_property_map.hpp>
#include <boost/property_map/vector_property_map.hpp>

namespace graph_model
{

  namespace graph
  {

    /// \brief Make a one-dimensional lattice (grid) graph
    ///
    /// Vertices are ordered from left to right across columns.
    ///
    template < typename Length >
    auto make_lattice (Length length)
    {
      assert(length > 0);
      using lattice_type = boost::grid_graph<1,long,long>;
      using array_type = boost::array<long,1>;
      array_type dimensions = { length };
      return lattice_type(dimensions);
    }

    /// \brief Make a two-dimensional lattice (grid) graph
    ///
    /// Vertices are ordered from upper left to lower right across
    /// columns and then along rows.
    ///
    template < typename Width, typename Height >
    auto make_lattice (Width width, Height height)
    {
      using lattice_type = boost::grid_graph<2,long,long>;
      using array_type = boost::array<long,2>;
      assert(width > 1 && height > 1);
      array_type dimensions = { width, height };
      return lattice_type(dimensions);
    }

    /// \brief Make a property map for a lattice
    ///
    /// Note that storage for the property map is \em not
    /// pre-allocated.  Consequently, references and iterators into
    /// the property map are \em not stable.
    ///
    template < typename T >
    auto make_lattice_properties ()
    {
      static_assert(std::is_default_constructible<T>::value,
		    "T must be default constructible");
      static_assert(std::is_copy_constructible<T>::value,
		    "T must be copy constructible");
      using property_map_type = boost::vector_property_map<T>;
      return property_map_type();
    }

    /// \brief Make a property map for a lattice
    ///
    /// Storage for the property map is pre-allocated.  No memory
    /// allocation will be performed after construction, unless
    /// vertices outside the expected range are accessed.  Therefore,
    /// references and iterators into the property map should be
    /// stable.
    ///
    template < typename T, typename Length >
    auto make_lattice_properties (Length length)
    {
      static_assert(std::is_default_constructible<T>::value,
		    "T must be default constructible");
      using index_property_map_type = boost::typed_identity_property_map<long>;
      using property_map_type = boost::shared_array_property_map<T,index_property_map_type>;
      return property_map_type(length);
    }

    /// \brief Make a property map for a lattice
    ///
    /// Storage for the property map is pre-allocated.  No memory
    /// allocation will be performed after construction, unless
    /// vertices outside the expected range are accessed.  Therefore,
    /// references and iterators into the property map should be
    /// stable.
    ///
    template < typename Length, typename T >
    auto make_lattice_properties (Length length, T t)
    {
      static_assert(std::is_default_constructible<T>::value,
		    "T must be default constructible");
      static_assert(std::is_copy_constructible<T>::value,
		    "T must be copy constructible");
      using property_map_type = boost::vector_property_map<T>;
      using std::fill;
      property_map_type property_map(length);
      fill(property_map.storage_begin(),property_map.storage_end(),t);
      return property_map;
    }

    /// \brief Make a property map for lattice vertices
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of vertices in the lattice.  No memory allocation will be
    /// performed after construction, unless vertices outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional vertices).  Therefore, references and iterators
    /// into the property map should always be stable.
    ///
    template < typename T, std::size_t Dimensions, typename VertexIndex, typename EdgeIndex >
    auto make_lattice_vertex_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice)
    { return make_lattice_properties<T>(num_vertices(lattice)); }

    /// \brief Make a property map for lattice vertices
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of vertices in the lattice.  No memory allocation will be
    /// performed after construction, unless vertices outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional vertices).  Therefore, references and iterators
    /// into the property map should always be stable.
    ///
    template < std::size_t Dimensions, typename VertexIndex, typename EdgeIndex, typename T >
    auto make_lattice_vertex_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice, T t)
    { return make_lattice_properties(num_vertices(lattice),t); }

    /// \brief Make a property map for lattice vertices
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of vertices in the lattice.  No memory allocation will be
    /// performed after construction, unless vertices outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional vertices).  Therefore, references and iterators
    /// into the property map should always be stable.
    ///
    template < typename T, std::size_t Dimensions, typename VertexIndex, typename EdgeIndex, typename U >
    auto make_lattice_vertex_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice, U u)
    { return make_lattice_vertex_properties(lattice,T(u)); }

    /// \brief Make a property map for lattice edges
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of edges in the lattice.  No memory allocation will be
    /// performed after construction, unless edges outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional edges).  Therefore, references and iterators into
    /// the property map should always be stable.
    ///
    template < typename T, std::size_t Dimensions, typename VertexIndex, typename EdgeIndex >
    auto make_lattice_edge_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice)
    { return make_lattice_properties<T>(num_edges(lattice)); }

    /// \brief Make a property map for lattice edges
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of edges in the lattice.  No memory allocation will be
    /// performed after construction, unless edges outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional edges).  Therefore, references and iterators into
    /// the property map should always be stable.
    ///
    template < std::size_t Dimensions, typename VertexIndex, typename EdgeIndex, typename T >
    auto make_lattice_edge_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice, T t)
    { return make_lattice_properties(num_edges(lattice),t); }

    /// \brief Make a property map for lattice edges
    ///
    /// Storage for the property map is pre-allocated for the number
    /// of edges in the lattice.  No memory allocation will be
    /// performed after construction, unless edges outside the
    /// expected range are accessed (which is a programming error,
    /// because the lattice itself is immutable and will not have
    /// additional edges).  Therefore, references and iterators into
    /// the property map should always be stable.
    ///
    template < typename T, std::size_t Dimensions, typename VertexIndex, typename EdgeIndex, typename U >
    auto make_lattice_edge_properties (boost::grid_graph<Dimensions,VertexIndex,EdgeIndex> lattice, U u)
    { return make_lattice_edge_properties(lattice,T(u)); }

  } // namespace graph

} // namespace graph_model

#endif
