/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_SIGNALS_SIGNAL_HPP
#define GRAPH_MODEL_SIGNALS_SIGNAL_HPP 1

#include <graph_model/signals/signal_fwd.hpp>
#include <type_traits>
#include <graph_model/signals/type/combiner.hpp>
#include <graph_model/type/slot_collection.hpp>
#include <graph_model/signals/type/invocation_policy.hpp>
#include <graph_model/utility/type/invocation_policy.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace signals
  {

    template < typename Result
	       , typename ... Args
	       , typename Combiner
	       , typename SignalTag
	       , template < typename ... > class SlotCollection
	       >
    class signal< Result(Args...),Combiner,SignalTag,SlotCollection >
    {
      using self = signal;
      using slot_collection_type = graph_model::type::slot_collection_t<SlotCollection,Result(Args...)>;
      using slot_iterator = typename slot_collection_type::const_iterator;

      template < typename C, typename Type = type::combiner_t<Combiner> >
      struct CombinerType
      {
	using type = Type;

	CombinerType () = default;
	explicit CombinerType (Type combiner) : combiner_(combiner) {}

	auto combiner () const { return combiner_; }

	void operator () () { combiner_(); }

	template < typename T >
	void operator () (T t) { combiner_(t); }

      private:
	Type combiner_;
      };

      template < typename C >
      struct CombinerType < C,void >
      {
	using type = void;

	CombinerType () = default;
	template < typename T >
	CombinerType (T) {}

	void operator () () {}

	template < typename T >
	void operator () (T) {}
      };

      using combiner_type = CombinerType<Combiner>;
      using invocation_policy_function_type = type::invocation_policy_t<combiner_type,slot_iterator,Args...>;
      using invocation_policy_type = utility::type::invocation_policy_t<use_default<>>;

    public:
      using size_type = typename slot_collection_type::size_type;

    public:
      signal () = default;
      template < typename C
		 , typename = mpl::enable_constructor_t<self,C>
		 >
      explicit signal (C combiner) : combiner_(combiner) {}

      auto& reset_invocation_strategy ()
      {
	invoke_ = invocation_policy_type();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_invocation_strategy (InvocationPolicy&& invocation_policy)
      {
	invoke_ = std::forward<InvocationPolicy>(invocation_policy);
	return *this;
      }

      size_type num_slots () const { return slots_.size(); }

      template < typename Slot >
      void connect (Slot&& slot)
      { slots_.insert(slots_.end(),std::forward<Slot>(slot)); }

      auto operator () (Args ... args)
      {
	using combiner_is_void = typename std::is_void<typename combiner_type::type>::type;
	return invoke(combiner_is_void(),args...);
      }

      auto operator () (Args ... args) const
      {
	using combiner_is_void = typename std::is_void<typename combiner_type::type>::type;
	return invoke(combiner_is_void(),args...);
      }

    private:
      auto invoke (std::false_type, Args ... args)
      { return invoke_(combiner_,slots_.begin(),slots_.end(),args...).combiner(); }

      auto invoke (std::false_type, Args ... args) const
      { return invoke_(combiner_,slots_.begin(),slots_.end(),args...).combiner(); }

    private:
      auto invoke (std::true_type, Args ... args)
      { return invoke_(slots_.begin(),slots_.end(),args...); }

      auto invoke (std::true_type, Args ... args) const
      { return invoke_(slots_.begin(),slots_.end(),args...); }

    private:
      slot_collection_type slots_;
      invocation_policy_function_type invoke_ = invocation_policy_type();
      combiner_type combiner_;
    };

    template < typename Result
	       , typename ... Args
	       , typename SignalTag
	       , template < typename ... > class SlotCollection
	       >
    class signal< Result(Args...),void,SignalTag,SlotCollection >
    {
      using slot_collection_type = graph_model::type::slot_collection_t<SlotCollection,Result(Args...)>;
      using slot_iterator = typename slot_collection_type::const_iterator;
      using invocation_policy_function_type = type::invocation_policy_t<void,slot_iterator,Args...>;
      using invocation_policy_type = utility::type::invocation_policy_t<use_default<>>;

    public:
      using size_type = typename slot_collection_type::size_type;

    public:
      auto& reset_invocation_strategy ()
      {
	invoke_ = invocation_policy_type();
	return *this;
      }

      template < typename InvocationPolicy >
      auto& reset_invocation_strategy (InvocationPolicy&& invocation_policy)
      {
	invoke_ = std::forward<InvocationPolicy>(invocation_policy);
	return *this;
      }

      size_type num_slots () const { return slots_.size(); }

      template < typename Slot >
      void connect (Slot&& slot) { slots_.insert(std::forward<Slot>(slot)); }

      auto operator () (Args ... args)
      { return invoke_(this->begin(),this->end(),args...); }

      auto operator () (Args ... args) const
      { return invoke_(this->begin(),this->end(),args...); }

    private:
      slot_collection_type slots_;
      invocation_policy_function_type invoke_ = invocation_policy_type();
    };

  } // namespace signals

} // namespace graph_model

#endif
