/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef OBSERVER_PTR_HPP
#define OBSERVER_PTR_HPP 1

#include <iostream>
#include <memory>
#include <type_traits>
#include <boost/mpl/if.hpp>
#include <boost/operators.hpp>
#include <always_false.hpp>
#include <mpl/enable_constructor.hpp>
#include <type_traits/is_rvalue.hpp>

#include <debugging/stack_trace.hpp>

/// \brief Mutable observer pointer
///
/// \b Ownership \b semantics: An observer pointer does \e not own its
/// pointee; it just provides access to (observes) it.  Note that the
/// lifetime of the pointee must exceed the last time the observer
/// pointer is dereferenced.
///
template < typename T >
class observer_ptr
  : boost::totally_ordered<observer_ptr<T>>
{
  template < typename U > friend class observer_ptr;

  using self = observer_ptr;
  
  using value_type = typename std::remove_reference<T>::type;
  using reference = value_type&;

  template < typename U >
  using is_constructible_from = std::enable_if_t<std::is_convertible<U*,T*>::value>;

 public:
  using element_type = T;
  using pointer = T*;
  
public:
  observer_ptr () = default;
  observer_ptr (self const&) = default;
  observer_ptr (self&&) = default;
  self& operator = (self const&) = default;
  self& operator = (self&&) = default;
  ~observer_ptr () { reset(); }

  observer_ptr (T* p) : p_(p) {}

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  observer_ptr (observer_ptr<U> other) : observer_ptr(other.p_) {}

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  observer_ptr (U* p) : p_(p) {}

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  observer_ptr (U& u) : p_(&u) {}

  observer_ptr (std::unique_ptr<T> const& p) : p_(p.get()) {}

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  observer_ptr (std::unique_ptr<U> const& p) : p_(p.get()) {}

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  self& operator = (observer_ptr<U> other) { swap(other); return *this; }

 public:
  /// \brief Dereference the pointer
  /// \pre The pointee exists.
  reference operator * () const { is_ok(); return *p_; }

  /// \brief Dereference the pointer
  /// \pre The pointee exists
  pointer operator -> () const { is_ok(); return p_; }

  template < typename U
	     , typename = is_constructible_from<U>
	     >
  void swap (observer_ptr<U>& other) noexcept
  { swap(static_cast<pointer>(other.p_)); }

  void swap (self& other) noexcept { swap(other.p_); }
  void swap (pointer other) noexcept { using std::swap; swap(p_,other); }
  
  void reset () noexcept { p_ = nullptr; }

  explicit operator bool () const { return static_cast<bool>(p_); }

  friend bool operator == (self lhs, self rhs) { return lhs.p_ == rhs.p_; }
  friend bool operator <  (self lhs, self rhs) { return lhs.p_ <  rhs.p_; }

  pointer get () const { return p_; }

 private:
  void is_ok () const
  {
    if (!p_)
      {
	std::cout << __FILE__ << ":" << __LINE__ << ":  ERROR: observer_ptr is null" << std::endl;
	std::cout << debugging::stack_trace<>() << std::endl;
	assert(p_);
      }
  }

private:
  pointer p_ = nullptr;
};

template < typename T >
void swap (observer_ptr<T>& left, observer_ptr<T>& right)
{ left.swap(right); }

template < typename T >
void swap (observer_ptr<T>& left, T* right)
{ left.swap(right); }

template < typename T >
void swap (T* left, observer_ptr<T>& right)
{ swap(right,left); }

template < typename T >
auto make_observer (observer_ptr<T> p) { return p; }

template < typename T >
auto make_observer (T& t) { return observer_ptr<T>(t); }

template < typename T >
auto make_observer (T const& t) { return observer_ptr<T const>(t); }

template < typename T
	   , typename = std::enable_if_t<type_traits::is_rvalue_v<T>>
           >
auto make_observer (T&& t)
{
  static_assert(always_false_v<T>,
		"An observer_ptr cannot be constructed from an rvalue");
}

template < typename T >
auto make_observer (std::unique_ptr<T> const& p)
{ return make_observer(*p.get()); }

template < typename T >
auto make_observer (std::unique_ptr<T>& p)
{ return make_observer(*p.get()); }

template < typename T, typename U >
auto make_observer (U const& u) { return make_observer(static_cast<T&>(u)); }

template < typename T
	   , typename U
	   , typename = std::enable_if_t<type_traits::is_rvalue_v<U>>
	   >
auto make_observer (U&& u)
{
  static_assert
    (always_false_v<T,U>,
     "An observer_ptr cannot be constructed from an rvalue");
}

#endif
