/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE sample_space_math_multinomial_pdf

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <sample_space/math/multinomial_pdf.hpp>
#include <sample_space/vector/simplex.hpp>
#include <sample_space/vector/vector.hpp>
#include <traits/vector.hpp>
#include <boost/math/special_functions/binomial.hpp>
#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<
  boost::mpl::pair<std::vector<float>,std::vector<int>>
  , boost::mpl::pair<std::vector<float>,std::vector<long>>
  , boost::mpl::pair<std::vector<double>,std::vector<int>>
  , boost::mpl::pair<std::vector<double>,std::vector<long>>
  , boost::mpl::pair<std::vector<long double>,std::vector<long long>>
  , boost::mpl::pair<sample_space::vector::simplex<float>,sample_space::vector::vector<int>>
  , boost::mpl::pair<sample_space::vector::simplex<double>,sample_space::vector::vector<long>>
  , boost::mpl::pair<sample_space::vector::simplex<long double>,sample_space::vector::vector<long long>>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_multinomial_pdf,Pair,test_types)
{
  using probability_vector_type = typename Pair::first;
  using probability_type = typename probability_vector_type::value_type;
  using size_vector_type = typename Pair::second;
  using size_type = typename size_vector_type::value_type;

  using std::abs;
  using std::pow;
  using boost::math::binomial_coefficient;
  using sample_space::math::make_multinomial_pdf;
  using sample_space::math::make_multinomial_coefficient;
  using sample_space::math::make_categorical_pdf;

  probability_vector_type p = { 0.1, 0.2, 0.3, 0.4 };
  size_vector_type n = { 1, 2, 3, 4 };

  auto expected_multinomial_coefficient
    = binomial_coefficient<probability_type>(10,1)
    * binomial_coefficient<probability_type>(9,2)
    * binomial_coefficient<probability_type>(7,3)
    * binomial_coefficient<probability_type>(4,4);
  auto expected_categorical_pdf
    = pow(p[0],n[0])
    * pow(p[1],n[1])
    * pow(p[2],n[2])
    * pow(p[3],n[3]);
  probability_type expected
    = expected_multinomial_coefficient * expected_categorical_pdf;

  auto multinomial_coefficient = make_multinomial_coefficient<probability_type>(n);
  BOOST_CHECK_EQUAL(multinomial_coefficient.value(),expected_multinomial_coefficient);

  auto categorical_pdf = make_categorical_pdf(p,n);
  BOOST_CHECK_EQUAL(categorical_pdf.value(),expected_categorical_pdf);

  auto multinomial_pdf = make_multinomial_pdf<probability_type>(p,n);
  BOOST_CHECK_EQUAL(multinomial_pdf.value(),expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_multinomial_pdf_ratio,Pair,test_types)
{
  using probability_vector_type = typename Pair::first;
  using probability_type = typename probability_vector_type::value_type;
  using size_vector_type = typename Pair::second;
  using size_type = typename size_vector_type::value_type;

  using std::abs;
  using std::pow;
  using boost::math::binomial_coefficient;
  using sample_space::math::make_multinomial_pdf;
  
  probability_vector_type p_1 = { 0.1, 0.2, 0.3, 0.4 };
  size_vector_type n_1 = { 1, 2, 3, 4 };

  probability_vector_type p_2 = { 0.1, 0.2, 0.3, 0.4 };
  size_vector_type n_2 = { 4, 3, 2, 1 };

  auto expected_multinomial_coefficient_1
    = binomial_coefficient<probability_type>(10,1)
    * binomial_coefficient<probability_type>(9,2)
    * binomial_coefficient<probability_type>(7,3)
    * binomial_coefficient<probability_type>(4,4);
  auto expected_categorical_pdf_1
    = pow(p_1[0],n_1[0])
    * pow(p_1[1],n_1[1])
    * pow(p_1[2],n_1[2])
    * pow(p_1[3],n_1[3]);
  probability_type expected_1
    = expected_multinomial_coefficient_1 * expected_categorical_pdf_1;

  auto expected_multinomial_coefficient_2
    = binomial_coefficient<probability_type>(10,4)
    * binomial_coefficient<probability_type>(6,3)
    * binomial_coefficient<probability_type>(3,2)
    * binomial_coefficient<probability_type>(1,1);
  auto expected_categorical_pdf_2
    = pow(p_2[0],n_2[0])
    * pow(p_2[1],n_2[1])
    * pow(p_2[2],n_2[2])
    * pow(p_2[3],n_2[3]);
  probability_type expected_2
    = expected_multinomial_coefficient_2 * expected_categorical_pdf_2;

  probability_type expected = expected_1 / expected_2;
  
  auto multinomial_pdf_ratio
    = make_multinomial_pdf<probability_type>(p_1,n_1)
    / make_multinomial_pdf<probability_type>(p_2,n_2);
  BOOST_CHECK_EQUAL(multinomial_pdf_ratio.value(),expected);
}
